/***
 * Juliac
 * Copyright (C) 2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.julia.control.binding;

import java.util.Map;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.api.control.SuperController;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.julia.Controller;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.juliac.runtime.RuntimeException;
import org.objectweb.fractal.koch.factory.MBindingControlComponent;
import org.objectweb.fractal.runtime.ultracomp.UltraCompUtils;
import org.objectweb.fractal.util.ContentControllerHelper;
import org.objectweb.fractal.util.Fractal;

/**
 * Provides a container based implementation of the {@link BindingController}
 * interface. This mixin implements the {@link BindingController} methods
 * through delegation to the required content control-component. This class acts
 * as a replacement for ContainerBindingControllerMixin in the ULTRA_COMP mode.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.5
 */
public abstract class UltraCompContainerBindingControllerMixin
implements Controller, BindingController, MBindingControlComponent {

    // -------------------------------------------------------------------------
    // Implementation of the MBindingControlComponent interface
    // -------------------------------------------------------------------------

	public void setFcOwner( Component owner ) {
		fcOwner = owner;
	}
	private Component fcOwner;

	
    // -------------------------------------------------------------------------
    // Fields and methods added and overriden by the mixin class
    // -------------------------------------------------------------------------

    private Map<String,Object> interfaces;
		
    /**
     * Initializes the fields of this mixin and then calls the overriden method.
     *
     * @param ic information about the component to which this controller object
     *      belongs.
     * @throws InstantiationException if the initialization fails.
     */

    public void initFcController (final InitializationContext ic)
    throws InstantiationException {
    	
	    this.interfaces = ic.interfaces;
	    
	    Component owner = (Component) ic.getOptionalInterface("component");
	    BindingController content =
    		(BindingController) interfaces.get("content");
	    try {
		    content.bindFc("component",owner);
	    }
	    catch( Exception ignored ) {}

	    _super_initFcController(ic);
    }

    public String[] listFc() {
	    BindingController contentitf = null;
	    try {
			contentitf = (BindingController) interfaces.get("content");
		    return contentitf.listFc();
	    }
	    catch( NullPointerException ignored ) {
			  /*
			   * The content component does not necessarily implements the
			   * BindingController interface. In such a case, the chain
			   * binding control component > component-based control membrane
			   * ends with a NPE.
			   */
	    }
	    return new String[0];
    }

    public Object lookupFc( String clientItfName )
    throws NoSuchInterfaceException {
    	
	    BindingController contentitf =
		    (BindingController) interfaces.get("content");
	    return contentitf.lookupFc(clientItfName);
    }

    public void bindFc( String clientItfName, Object serverItf)
	throws
		NoSuchInterfaceException, IllegalBindingException,
		IllegalLifeCycleException {
    	
    	/*
    	 * Traverse the component hierarchy down to the mComposite that hosts
    	 * the component-based membrane, the interceptor components and the
    	 * content component.
    	 */
    	Component toplevel = getTopLevelComposite();
    	
    	/*
    	 * Retrieve the interface to be bound to the content component. This is
    	 * either one from the interceptor component (if there is one), or the
    	 * an internal one from the mComposite.
    	 */
    	Object itf = null;
		String icname =
			UltraCompUtils.INTERCEPTOR_COMPONENT_PREFIX +
			UltraCompUtils.toIdentifier(clientItfName);
    	Component ic =
			ContentControllerHelper.getSubComponentByName(toplevel,icname);
    	if( ic == null ) {
    		// No interceptor
        	ContentController cc = Fractal.getContentController(toplevel);
        	itf = (Interface) cc.getFcInternalInterface(clientItfName);
    	}
    	else {
    		itf = ic.getFcInterface(clientItfName);
    	}
    	
    	/*
    	 * Bind the content component.
    	 */
    	Component content =
			ContentControllerHelper.getSubComponentByName(
				toplevel, UltraCompUtils.CONTENT_COMPONENT_NAME );
    	BindingController bc = Fractal.getBindingController(content);
    	bc.bindFc(clientItfName,itf);
    	
    	/*
    	 * Bind the mComposite.
    	 */
    	BindingController toplevelbc = Fractal.getBindingController(toplevel);
    	toplevelbc.bindFc(clientItfName,serverItf);
    }

    public void unbindFc( String clientItfName )
	throws
		NoSuchInterfaceException, IllegalBindingException,
		IllegalLifeCycleException {
	  
    	/*
    	 * Traverse the component hierarchy down to the mComposite that hosts
    	 * the component-based membrane, the interceptor components and the
    	 * content component.
    	 */
	    Component toplevel = getTopLevelComposite();

    	/*
    	 * Unbind the content component.
    	 */
	    Component content =
			ContentControllerHelper.getSubComponentByName(
				toplevel, UltraCompUtils.CONTENT_COMPONENT_NAME );
        BindingController bc = Fractal.getBindingController(content);
        bc.unbindFc(clientItfName);
        
    	/*
    	 * Unbind the mComposite.
    	 */
    	BindingController toplevelbc = Fractal.getBindingController(toplevel);
    	toplevelbc.unbindFc(clientItfName);
    }

    
    // -------------------------------------------------------------------------
    // Implementation specific
    // -------------------------------------------------------------------------

    /**
     * Return the top level mComposite that holds the component-based control
     * membrane, the interceptor components (if any), and the content component
     * (if any.)
     */
    private Component getTopLevelComposite()
	throws NoSuchInterfaceException {

    	/*
    	 * Traverse the component hierarchy down to the mComposite that hosts
    	 * the component-based membrane, the interceptor components and the
    	 * content component.
    	 */
    	Component current = fcOwner;
    	NameController nc = Fractal.getNameController(current);
    	String name = nc.getFcName();
    	while( ! name.startsWith(UltraCompUtils.TOP_LEVEL_PREFIX) ) {
    		SuperController sc = Fractal.getSuperController(current);
    		Component[] supers = sc.getFcSuperComponents();
    		if( supers.length != 1 ) {
    			String msg =
					"Unsupported number of super components: "+supers.length;
    			throw new RuntimeException(msg);
    		}
    		current = supers[0];
        	nc = Fractal.getNameController(current);
        	name = nc.getFcName();
    	}
    	
    	return current;
    }
    
 
    // -------------------------------------------------------------------------
    // Fields and methods required by the mixin class in the base class
    // -------------------------------------------------------------------------

    /**
     * The {@link Controller#initFcController initFcController} method overriden
     * by this mixin.
     *
     * @param ic information about the component to which this controller object
     *      belongs.
     * @throws InstantiationException if the initialization fails.
     */
    public abstract void _super_initFcController( InitializationContext ic )
    throws InstantiationException;
}
