/***
 * Juliac
 * Copyright (C) 2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.julia.control.lifecycle;

import java.util.Map;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.julia.Controller;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.juliac.control.lifecycle.InvocationCounterItf;

/**
 * Provides container related functions to a {@link LifeCycleController}. This
 * class acts as a replacement for ContainerLifeCycleMixin in the ULTRA_COMP
 * mode.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.5
 */
public abstract class UltraCompContainerLifeCycleMixin
implements Controller, LifeCycleCoordinator, InvocationCounterItf {

	/*
	 * Implementing InvocationCounterItf here is a trick to keep the
	 * org.objectweb.fractal.julia.control.lifecycle.BasicLifeCycleControllerMixin
	 * class from Julia unchanged.
	 */ 
	
  // -------------------------------------------------------------------------
  // Fields and methods added and overriden by the mixin class
  // -------------------------------------------------------------------------

  private Map<String,Object> interfaces;
	
  /**
   * Initialize this controller object.
   *
   * @param ic information about the component to which this controller object
   *      belongs.
   * @throws InstantiationException if the initialization fails.
   */

  public void initFcController( InitializationContext ic )
  throws InstantiationException {
	  this.interfaces = ic.interfaces;
	  _super_initFcController(ic);
  }

  /**
   * Calls the overriden method and then calls the {@link #setFcContentState
   * setFcContentState} method.
   * 
   * @return <tt>true</tt> if the execution state has changed, or <tt>false</tt>
   *      if it had already the {@link #STARTED STARTED} value.
   * @throws IllegalLifeCycleException if a problem occurs.
   */

  public boolean setFcStarted () throws IllegalLifeCycleException {
    synchronized (this) {
      if (_super_setFcStarted()) {
        setFcContentState(true);
        return true;
      }
      return false;
    }
  }

  /**
   * Calls the overriden method and then calls the {@link #setFcContentState
   * setFcContentState} method.
   *
   * @return <tt>true</tt> if the execution state has changed, or <tt>false</tt>
   *      if it had already the {@link #STOPPED STOPPED} value.
   * @throws IllegalLifeCycleException if a problem occurs.
   */

  public boolean setFcStopped () throws IllegalLifeCycleException {
    synchronized (this) {
      if (_super_setFcStopped()) {
        setFcContentState(false);
        return true;
      }
      return false;
    }
  }

  /**
   * Calls the {@link LifeCycleController#startFc startFc} or {@link
   * LifeCycleController#stopFc stopFc} method on the encapsulated component.
   * This method does nothing if there is no encapsulated component, or if it
   * does not implement the {@link LifeCycleController} interface.
   *
   * @param started <tt>true</tt> to call the {@link LifeCycleController#startFc
   *      startFc}, or <tt>false</tt> to call the {@link
   *      LifeCycleController#stopFc stopFc} method.
   * @throws IllegalLifeCycleException if a problem occurs.
   */

  public void setFcContentState (final boolean started)
  throws IllegalLifeCycleException {
	  
	  LifeCycleController content =
		  (LifeCycleController) interfaces.get("content");
	  if( started ) {
		  try {
			  content.startFc();
		  }
		  catch( NullPointerException ignored ) {
			  /*
			   * The content component does not necessarily implements the
			   * LifeCycleController interface. In such a case, the chain
			   * lifeCycle control component > component-based control membrane
			   * ends with a NPE.
			   */
		  }
	  }
	  else {
		  try {
			  content.stopFc();
		  }
		  catch( NullPointerException ignored ) {}
	  }
  }

  // -------------------------------------------------------------------------
  // Fields and methods required by the mixin class in the base class
  // -------------------------------------------------------------------------

  /*
   * TODO the definition of the lifecycle control-component should
   * be updated to remove the corresponding required interface. The architecture
   * of the primitive control membrane should also be updated accordingly.
   */
  
//  /**
//   * The <tt>weaveableOptC</tt> field required by this mixin. This field is
//   * supposed to reference the {@link Component} interface of the component to
//   * which this controller object belongs.
//   */
//
//  public Component _this_weaveableOptC;

  /**
   * The {@link Controller#initFcController initFcController} method overriden
   * by this mixin.
   *
   * @param ic information about the component to which this controller object
   *      belongs.
   * @throws InstantiationException if the initialization fails.
   */

  public abstract void _super_initFcController (InitializationContext ic)
    throws InstantiationException;

  /**
   * The {@link LifeCycleCoordinator#setFcStarted setFcStarted} method overriden
   * by this mixin.
   * 
   * @return <tt>true</tt> if the execution state has changed, or <tt>false</tt>
   *      if it had already the {@link #STARTED STARTED} value.
   * @throws IllegalLifeCycleException if a problem occurs.
   */

  public abstract boolean _super_setFcStarted ()
    throws IllegalLifeCycleException;

  /**
   * The {@link LifeCycleCoordinator#setFcStopped setFcStopped} method overriden
   * by this mixin.
   *
   * @return <tt>true</tt> if the execution state has changed, or <tt>false</tt>
   *      if it had already the {@link #STOPPED STOPPED} value.
   * @throws IllegalLifeCycleException if a problem occurs.
   */

  public abstract boolean _super_setFcStopped ()
    throws IllegalLifeCycleException;
}
